import pysvgchart as psc
import random
import os
import datetime as dt
import math

random.seed(42)


def write_out(content, name, output_dir="showcase"):
    output_file = os.path.join(output_dir, name)
    os.makedirs(output_dir, exist_ok=True)
    with open(output_file, 'w+') as out_file:
        out_file.write(content)
    return output_file


def test_write_out_styles():
    write_out(psc.render_all_styles(), 'pysvgchart.css', 'styles')


def test_line_chart():
    x_values = list(range(100))
    y_values = [4000]
    for i in range(99):
        y_values.append(y_values[-1] + 100 * random.randint(0, 1))

    line_chart = psc.LineChart(
        x_values=x_values,
        y_values=[y_values, [1000 + y for y in y_values]],
        y_names=['predicted', 'actual'],
        x_max_ticks=20,
        y_zero=True,
        width=900,
        right_margin=200,
        y_margin=150
    )
    line_chart.add_grids(minor_y_ticks=4, minor_x_ticks=4)
    line_chart.add_legend()

    output_file = write_out(line_chart.render(), name="line.svg")

    assert os.path.exists(output_file), "SVG file was not created."
    assert 'svg' in line_chart.render().lower(), "SVG content is not in the render output."
    assert len(line_chart.y_axis.tick_texts) > 0, "Y-axis ticks are missing."
    assert line_chart.y_axis.tick_texts[-1].styles, "Y-axis tick styles are missing."
    assert isinstance(line_chart.series['predicted'].path_length, float), "Path length error"


STYLISED_LINE_CHART_DATA: dict | None = None


def init_stylised_line_chart_data():
    """
    make sure that the stylised line charts use the same data
    """
    global STYLISED_LINE_CHART_DATA
    if STYLISED_LINE_CHART_DATA is None:
        start_date = dt.date(2025, 2, 14)
        dates = [start_date - dt.timedelta(days=i) for i in range(500) if (start_date + dt.timedelta(days=i)).weekday() == 0][::-1]
        actual = [(1 + math.sin(d.timetuple().tm_yday / 183 * math.pi)) * 50000 + 1000 * i + random.randint(0, 20000) for i, d in enumerate(dates)]
        expected = [a + random.randint(-10000, 10000) for a in actual]
        STYLISED_LINE_CHART_DATA = {
            "dates": dates,
            "actual": actual,
            "expected": expected,
        }


def stylised_line_chart(x_shift=False, y_shift=False, x_log=False, y_log=False):
    def y_labels(num):
        num = float('{:.3g}'.format(num))
        magnitude = 0
        while abs(num) >= 1000:
            magnitude += 1
            num /= 1000.0
        rtn = '{}{}'.format('{:f}'.format(num).rstrip('0').rstrip('.'), ['', 'K', 'M', 'B', 'T'][magnitude])
        return rtn

    def x_labels(date):
        return date.strftime('%b')

    init_stylised_line_chart_data()
    dates = STYLISED_LINE_CHART_DATA["dates"]
    actual = [v if v > 0.0 or not y_log else 0.001 for v in STYLISED_LINE_CHART_DATA["actual"]]
    expected = [v if v > 0.0 or not y_log else 0.001 for v in STYLISED_LINE_CHART_DATA["expected"]]
    chart = psc.LineChart(
        x_values=dates,
        y_values=[actual, expected],
        y_names=['Actual sales', 'Predicted sales'],
        x_max_ticks=30,
        x_label_format=x_labels,
        y_label_format=y_labels,
        width=1200,
        left_margin=100,
        right_margin=100,
        x_shift=x_shift,
        y_shift=y_shift,
        x_log=x_log,
        y_log=y_log,
    )
    chart.series['Actual sales'].styles = {'stroke': "#DB7D33", 'stroke-width': '3'}
    chart.series['Predicted sales'].styles = {'stroke': '#2D2D2D', 'stroke-width': '3', 'stroke-dasharray': '4,4'}
    chart.add_legend(x_position=700, element_x=200, element_y=0, y_position=60, line_length=35, line_text_gap=20)
    chart.add_y_grid(minor_ticks=0, major_grid_style={'stroke': '#E9E9DE'})
    chart.x_axis.tick_lines, chart.y_axis.tick_lines = [], []
    chart.x_axis.axis_line = None
    chart.y_axis.axis_line.styles['stroke'] = '#E9E9DE'
    line_end = chart.legend.lines[0].end
    act_styles = {'fill': '#FFFFFF', 'stroke': '#DB7D33', 'stroke-width': '3'}
    chart.add_custom_element(psc.Circle(x=line_end.x, y=line_end.y, radius=4, styles=act_styles))
    line_end = chart.legend.lines[1].end
    pred_styles = {'fill': '#2D2D2D', 'stroke': '#2D2D2D', 'stroke-width': '3'}
    chart.add_custom_element(psc.Circle(x=line_end.x, y=line_end.y, radius=4, styles=pred_styles))
    for limit, tick in zip(chart.x_axis.scale.ticks, chart.x_axis.tick_texts):
        if tick.content == 'Jan':
            chart.add_custom_element(psc.Text(x=tick.position.x, y=tick.position.y + 15, content=str(limit.year), styles=tick.styles))

    def hover_modifier(position, x_value, y_value, series_name, styles=None):
        default_stroke = "#808080"
        classes = [
            psc.hover_style_name,
        ]
        marker_styles = {
            "fill": "#FFFFFF",
            "stroke": default_stroke if styles is None else styles.get("stroke", default_stroke),
            "stroke-width": "3",
        }
        text_styles = {
            "alignment-baseline": "middle",
            "text-anchor": "middle",
        }
        x_content = str(x_value)
        y_content = "{:,.0f}".format(y_value)
        return [
            psc.Circle(x=position.x, y=position.y, radius=3, classes=classes, styles=marker_styles),
            psc.Text(x=position.x, y=position.y - 10, content=x_content, classes=classes, styles=text_styles),
            psc.Text(x=position.x, y=position.y - 30, content=y_content, classes=classes, styles=text_styles),
            psc.Text(x=position.x, y=position.y - 50, content=series_name, classes=classes, styles=text_styles)
        ]

    chart.add_hover_modifier(hover_modifier, radius=3)
    return chart


def test_stylised_line_chart_without_shift():
    chart = stylised_line_chart(x_shift=False, y_shift=False)
    write_out(chart.render_with_all_styles(), name="detailed.svg")


def test_stylised_line_chart_with_x_shift():
    chart = stylised_line_chart(x_shift=True, y_shift=False)
    write_out(chart.render_with_all_styles(), name="detailed-x-shift.svg")


def test_stylised_line_chart_with_y_shift():
    chart = stylised_line_chart(x_shift=False, y_shift=True)
    write_out(chart.render_with_all_styles(), name="detailed-y-shift.svg")


def test_stylised_line_chart_with_xy_shift():
    chart = stylised_line_chart(x_shift=True, y_shift=True)
    write_out(chart.render_with_all_styles(), name="detailed-xy-shift.svg")


def test_stylised_log_chart_with_y_log():
    chart = stylised_line_chart(x_log=False, y_log=True)
    write_out(chart.render_with_all_styles(), name="detailed-y-log.svg")


def test_donut():
    values = [11.3, 20, 30, 40]
    donut_chart = psc.DonutChart(values)
    write_out(donut_chart.render(), name="donut.svg")


def test_donut_legend():
    values = [11.3, 20, 30, 40]
    donut_chart = psc.DonutChart(values)
    donut_chart.add_legend()
    write_out(donut_chart.render(), name="donut_legend.svg")


def test_donut_hover():
    def hover_modifier(position, name, value, chart_total):
        text_styles = {'alignment-baseline': 'middle', 'text-anchor': 'middle'}
        params = {'styles': text_styles, 'classes': [psc.hover_style_name]}
        return [
            psc.Text(x=position.x, y=position.y - 10, content=name, **params),
            psc.Text(x=position.x, y=position.y + 10, content="{:.2%}".format(value / chart_total), **params)
        ]

    values = [11.3, 20, 30, 40]
    names = ['Apples', 'Bananas', 'Cherries', 'Durians']
    donut_chart = psc.DonutChart(values, names)
    donut_chart.add_hover_modifier(hover_modifier)
    write_out(donut_chart.render_with_all_styles(), name="donut_hover.svg")


def test_simple_line_chart():
    values = [11.3, 20, 30, 40]
    names = ['Apples', 'Bananas', 'Cherries', 'Durians']
    simple_line_chart = psc.SimpleLineChart(x_values=names, y_values=[values], y_names=['number'], y_zero=True)
    write_out(simple_line_chart.render(), name="simple.svg")


def test_simple_line_chart_with_axis_titles():
    values = [11.3, 20, 30, 40]
    names = ['Apples', 'Bananas', 'Cherries', 'Durians']
    simple_line_chart = psc.SimpleLineChart(x_values=names, y_values=[values], y_names=['number'], y_zero=True, x_axis_title='Fruit', y_axis_title='Sold')
    write_out(simple_line_chart.render(), name="simple_titled.svg")

def test_simple_line_chart_with_secondary_axis_titles():
    values = [11.3, 20, 30, 40]
    names = ['Apples', 'Bananas', 'Cherries', 'Durians']
    simple_line_chart = psc.SimpleLineChart(x_values=names, y_values=[values], sec_y_values=[values], sec_y_names=['number'], sec_y_zero=True, x_axis_title='Fruit', sec_y_axis_title='Sold')
    write_out(simple_line_chart.render(), name="simple_titled_secondary.svg")


def test_bar_chart():
    values = [[10, 20, 30, 40], [30, 10, 10, 20]]
    names = ['Apples', 'Bananas', 'Cherries', 'Durians']
    bar_chart = psc.BarChart(x_values=names, y_values=values, y_names=['Monday', 'Tuesday'], y_zero=True, width=900, right_margin=200)
    bar_chart.add_legend()
    write_out(bar_chart.render(), name="bar.svg")


def test_normalised_bar_chart():
    values = [[10, 20, 30, 40], [30, 10, 10, 20]]
    names = ['Apples', 'Bananas', 'Cherries', 'Durians']
    bar_chart = psc.NormalisedBarChart(x_values=names, y_values=values, y_names=['Monday', 'Tuesday'], y_zero=True, width=900, right_margin=200, y_label_format=lambda value: "{:.0%}".format(value))
    bar_chart.add_legend()
    write_out(bar_chart.render(), name="normalised_bar.svg")


SCATTER_CHART_DATA: dict | None = None


def init_scatter_chart_data():
    """
    make sure that the scatter charts use the same data
    """
    global SCATTER_CHART_DATA
    if SCATTER_CHART_DATA is None:
        x_values = [random.random() for _ in range(50)]
        y_values = [
            [x * 0.3 + random.random() * 0.7 for x in x_values],
            [x * 0.5 + random.random() * 0.5 for x in x_values],
        ]
        SCATTER_CHART_DATA = {
            "x_values": x_values,
            "y_values": y_values,
        }


def scatter_chart(x_shift=False, y_shift=False, x_log=False, y_log=False):
    init_scatter_chart_data()
    x_values = SCATTER_CHART_DATA["x_values"]
    y_values = SCATTER_CHART_DATA["y_values"]
    chart = psc.ScatterChart(x_values=x_values, y_values=y_values, y_names=["A", "B"], x_shift=x_shift, y_shift=y_shift, x_log=x_log, y_log=y_log)
    chart.add_legend()
    return chart


def test_scatter_chart_without_shift():
    chart = scatter_chart(x_shift=False, y_shift=False)
    write_out(chart.render(), name="scatter.svg")


def test_scatter_chart_xy_shift():
    chart = scatter_chart(x_shift=True, y_shift=True)
    write_out(chart.render(), name="scatter-xy-shift.svg")


def test_scatter_chart_x_log():
    chart = scatter_chart(x_log=True, y_log=False)
    write_out(chart.render(), name="scatter-x-log.svg")


def test_scatter_chart_y_log():
    chart = scatter_chart(x_log=False, y_log=True)
    write_out(chart.render(), name="scatter-y-log.svg")


def test_scatter_chart_xy_log():
    chart = scatter_chart(x_log=True, y_log=True)
    write_out(chart.render(), name="scatter-xy-log.svg")



from abc import ABC, abstractmethod
from dataclasses import dataclass

from .helpers import collapse_element_list
from .shared import number, style_def


@dataclass
class Point:
    """
    point in 2D space
    """

    x: number
    y: number


class Element(ABC):
    """
    abstract base class for all visual elements
    """

    __default_classes__: list[str] = []
    __default_styles__: style_def = {}

    def __init__(
        self,
        styles: style_def | None = None,
        classes: list[str] | None = None,
    ):
        self.styles = self.__default_styles__.copy() if styles is None else styles
        self.classes = self.__default_classes__.copy() if classes is None else classes

    @property
    def attributes(self) -> str:
        attributes = (
            {
                **self.styles,
                "class": " ".join(self.classes),
            }
            if len(self.classes) > 0
            else self.styles
        )
        return " ".join([a + '="' + attributes[a] + '"' for a in attributes])

    def add_classes(self, classes: list[str]) -> None:
        self.classes.extend(classes)

    @abstractmethod
    def get_element_list(self) -> list: ...


class Shape(Element):
    """
    abstract base class for all shapes
    """

    def __init__(
        self,
        x: number,
        y: number,
        styles: style_def | None = None,
        classes: list[str] | None = None,
    ):
        super().__init__(styles, classes)
        self.position = Point(x=x, y=y)

    @abstractmethod
    def get_element_list(self) -> list: ...


class Line(Shape):
    """
    straight line between two points
    """

    line_template = '<line x1="{x1}" y1="{y1}" x2="{x2}" y2="{y2}" {attributes}/>'

    def __init__(
        self,
        x: number,
        y: number,
        *,
        width: number | None = None,
        height: number | None = None,
        x2: number | None = None,
        y2: number | None = None,
        styles: style_def | None = None,
        classes: list[str] | None = None,
    ):
        if width is None and height is None and x2 is not None and y2 is not None:
            pass
        elif width is not None and height is not None and x2 is None and y2 is None:
            x2 = x + width
            y2 = y + height
        else:
            raise ValueError("use either width and height, or x2 and y2")
        super().__init__(x, y, styles, classes)
        self.end = Point(x2, y2)
        self.styles = dict() if styles is None else styles

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} start={self.position} end={self.end}>"

    @property
    def start(self) -> Point:
        return self.position

    def get_element_list(self) -> list:
        return [
            self.line_template.format(
                x1=self.start.x,
                y1=self.start.y,
                x2=self.end.x,
                y2=self.end.y,
                attributes=self.attributes,
            ),
        ]


class Circle(Shape):
    """
    circle around a center point
    """

    circle_template = '<circle cx="{x}" cy="{y}" r="{r}" {attributes}/>'

    def __init__(
        self,
        x: number,
        y: number,
        *,
        radius: number | None = None,
        styles: style_def | None = None,
        classes: list[str] | None = None,
    ):
        super().__init__(x, y, styles, classes)
        self.radius = radius

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} c={self.position} r={self.radius}>"

    def get_element_list(self) -> list[str]:
        return [
            self.circle_template.format(
                x=self.position.x,
                y=self.position.y,
                r=self.radius,
                attributes=self.attributes,
            ),
        ]


class Rect(Shape):
    """
    rectangle at a position with dimensions
    """

    rect_template = '<rect x="{x}" y="{y}" width="{width}" height="{height}" {attributes}/>'

    def __init__(
        self,
        x: number,
        y: number,
        *,
        width: number | None = None,
        height: number | None = None,
        x2: number | None = None,
        y2: number | None = None,
        styles: style_def | None = None,
        classes: list[str] | None = None,
    ):
        if width is not None and height is not None and x2 is None and y2 is None:
            pass
        elif width is None and height is None and x2 is not None and y2 is not None:
            width = max(x, x2) - min(x, x2)
            height = max(y, y2) - min(y, y2)
        else:
            raise ValueError("use either width and height, or x2 and y2")
        super().__init__(x, y, styles, classes)
        self.width = width
        self.height = height

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} pos={self.position} w={self.width} h={self.height}>"

    def get_element_list(self) -> list[str]:
        return [
            self.rect_template.format(
                x=self.position.x,
                y=self.position.y,
                width=self.width,
                height=self.height,
                attributes=self.attributes,
            ),
        ]


class Text(Shape):
    """
    text at a position
    """

    text_template = '<text x="{x}" y="{y}" {attributes}>{content}</text>'

    def __init__(self, x, y, content, styles=None, classes=None):
        super().__init__(x, y, styles, classes)
        self.content = content

    def __repr__(self):
        return f"<{self.__class__.__name__} pos={self.position} content={self.content} styles={self.styles}>"

    def get_element_list(self) -> list:
        return [
            self.text_template.format(
                x=self.position.x,
                y=self.position.y,
                content=self.content,
                attributes=self.attributes,
            ),
        ]


class Group(Element):
    """
    a group of visual elements
    """

    group_template = "<g {attributes}>"

    def __init__(
        self,
        styles=None,
        classes=None,
        children: list[Element] | None = None,
    ):
        super().__init__(styles, classes)
        self.children = [] if children is None else children

    def __repr__(self):
        return f"<{self.__class__.__name__} children={self.children}>"

    def add_children(self, children):
        self.children.extend(children)

    def get_element_list(self) -> list:
        return (
            [self.group_template.format(attributes=self.attributes)]
            + collapse_element_list(self.children)
            + ["</g>"]
        )



from __future__ import annotations
from abc import abstractmethod
from typing import Any, Callable

from .helpers import collapse_element_list
from .scales import make_linear_scale
from .shapes import Shape, Text, Line
from .shared import number, style_def


class Axis(Shape):
    """
    axis of a graph
    """

    default_axis_styles: style_def = {"stroke": "#2e2e2c"}

    def __init__(
        self,
        x_position: number,
        y_position: number,
        data_points,
        axis_length: number,
        label_format: Callable,
        max_ticks: int = 10,
        axis_styles: style_def | None = None,
        tick_length: int = 5,
        min_value=None,
        max_value=None,
        include_zero: bool = False,
        shift: bool = False,
        min_unique_values: int = 2,
        scale_maker=make_linear_scale,
        secondary: bool = False,
        title: str | None = None,
        title_styles: style_def | None = None,
        title_offset: int = 40
    ):
        _ignore = secondary, axis_styles, tick_length
        super().__init__(x_position, y_position)
        self.data_points = data_points
        self.length = axis_length
        self.scale = scale_maker(
            values=data_points,
            max_ticks=max_ticks,
            min_value=min_value,
            max_value=max_value,
            include_zero=include_zero,
            shift=shift,
            min_unique_values=min_unique_values,
        )
        self.label_format = label_format
        self.axis_line: Line | None = None
        self.tick_lines: list[Line] = []
        self.tick_texts: list[Text] = []
        self.grid_lines: list[Line] = []
        self.title: Text | None = None

    def get_element_list(self):
        return collapse_element_list(
            [self.axis_line, self.title] if self.title else [self.axis_line],
            self.tick_lines,
            self.tick_texts,
            self.grid_lines,
        )

    @abstractmethod
    def get_positions(self, values, include_axis=True) -> list[int | float | None]: ...

    def get_ticks_with_positions(self) -> list[tuple[Any, int | float | None]]:
        return list(zip(self.scale.ticks, self.get_positions(self.scale.ticks)))


class XAxis(Axis):
    """
    x-axis of a graph
    """

    default_tick_text_styles = {"text-anchor": "middle", "dominant-baseline": "hanging"}
    default_title_styles = {"text-anchor": "middle", "dominant-baseline": "middle"}

    def __init__(
        self,
        x_position: number,
        y_position: number,
        data_points,
        axis_length: number,
        label_format: Callable,
        max_ticks: int = 10,
        axis_styles: style_def | None = None,
        tick_length: int = 5,
        min_value=None,
        max_value=None,
        include_zero: bool = False,
        shift: bool = False,
        scale_maker=make_linear_scale,
        title: str | None = None,
        title_styles: style_def | None = None,
        title_offset: int = 40
    ):
        super().__init__(
            x_position=x_position,
            y_position=y_position,
            data_points=data_points,
            axis_length=axis_length,
            label_format=label_format,
            max_ticks=max_ticks,
            axis_styles=axis_styles,
            tick_length=tick_length,
            min_value=min_value,
            max_value=max_value,
            include_zero=include_zero,
            shift=shift,
            min_unique_values=2,  # at least two unique values needed on the x-axis to create a meaningful graph
            scale_maker=scale_maker,
        )
        styles = axis_styles or self.default_axis_styles.copy()
        self.axis_line = Line(
            x=self.position.x,
            y=self.position.y,
            width=axis_length,
            height=0,
            styles=styles,
        )

        for tick, pos in self.get_ticks_with_positions():
            if pos is None:  # shifted out of the visible range
                continue
            self.tick_lines.append(
                Line(x=pos, width=0, y=self.position.y, height=tick_length, styles=styles),
            )
            self.tick_texts.append(
                Text(
                    x=pos,
                    y=self.position.y + 2 * tick_length,
                    content=label_format(tick),
                    styles=self.default_tick_text_styles.copy(),
                ),
            )

        if title:
            title_x = self.position.x + self.length / 2  # Center on the axis
            title_y = self.position.y + title_offset
            self.title = Text(
                x=title_x,
                y=title_y,
                content=title,
                styles=title_styles or self.default_title_styles.copy()
            )

    def get_positions(self, values, include_axis=True) -> list[int | float | None]:
        proportions_of_range = [self.scale.value_to_fraction(value) for value in values]
        in_range = (
            (lambda prop: 0.0 <= prop <= 1.0) if include_axis else (lambda prop: 0.0 < prop <= 1.0)
        )
        return [
            self.position.x + prop * self.length if in_range(prop) else None
            for prop in proportions_of_range
        ]


class YAxis(Axis):
    """
    y-axis of a graph
    """

    default_tick_text_styles = {"text-anchor": "end", "dominant-baseline": "middle"}
    default_sec_tick_text_styles = {"text-anchor": "start", "dominant-baseline": "middle"}
    default_title_styles = {"text-anchor": "middle", "dominant-baseline": "middle"}

    def __init__(
        self,
        x_position: number,
        y_position: number,
        data_points,
        axis_length: number,
        label_format: Callable,
        max_ticks: int = 10,
        axis_styles: style_def | None = None,
        tick_length: int = 5,
        min_value=None,
        max_value=None,
        include_zero: bool = False,
        shift: bool = False,
        scale_maker=make_linear_scale,
        secondary: bool = False,
        title: str | None = None,
        title_styles: style_def | None = None,
        title_offset: int = 40
    ):
        super().__init__(
            x_position=x_position,
            y_position=y_position,
            data_points=data_points,
            axis_length=axis_length,
            label_format=label_format,
            max_ticks=max_ticks,
            axis_styles=axis_styles,
            tick_length=tick_length,
            min_value=min_value,
            max_value=max_value,
            include_zero=include_zero,
            shift=shift,
            secondary=secondary,
            min_unique_values=1,  # one unique value is sufficient for the y-axis
            scale_maker=scale_maker,
        )
        styles = axis_styles or self.default_axis_styles.copy()
        self.axis_line = Line(
            x=self.position.x,
            y=self.position.y,
            width=0,
            height=axis_length,
            styles=styles,
        )

        if secondary:
            tick_pos_offset = 0
            tick_text_offset = 2 * tick_length
            tick_text_styles = self.default_sec_tick_text_styles.copy()
        else:
            tick_pos_offset = -tick_length
            tick_text_offset = -2 * tick_length
            tick_text_styles = self.default_tick_text_styles.copy()

        for tick, pos in self.get_ticks_with_positions():
            if pos is None:  # shifted out of the visible range
                continue
            self.tick_lines.append(
                Line(
                    x=self.position.x + tick_pos_offset,
                    y=pos,
                    width=tick_length,
                    height=0,
                    styles=styles,
                ),
            )
            self.tick_texts.append(
                Text(
                    x=self.position.x + tick_text_offset,
                    y=pos,
                    content=label_format(tick),
                    styles=tick_text_styles,
                ),
            )

        if title:
            title_x = self.position.x + tick_text_offset + (title_offset if secondary else -title_offset)
            title_y = self.position.y + self.length / 2
            styles = title_styles or self.default_title_styles.copy()
            styles['transform'] = f'rotate(-90 {title_x} {title_y})'
            self.title = Text(
                x=title_x,
                y=title_y,
                content=title,
                styles=styles
            )

    def get_positions(self, values, include_axis=True) -> list[int | float | None]:
        proportions_of_range = [1 - self.scale.value_to_fraction(value) for value in values]
        in_range = (
            (lambda prop: 0.0 <= prop <= 1.0) if include_axis else (lambda prop: 0.0 <= prop < 1.0)
        )
        return [
            self.position.y + prop * self.length if in_range(prop) else None
            for prop in proportions_of_range
        ]


class CategoryYAxis(YAxis):
    """
    Y-axis for categorical data where categories should appear in order from top to bottom.
    Unlike the standard YAxis which inverts the scale (higher values at top),
    this axis preserves the category order (first category at top).
    """

    def get_positions(self, values, include_axis=True) -> list[int | float | None]:
        # Don't invert for categories - we want first category at top (y=0)
        proportions_of_range = [self.scale.value_to_fraction(value) for value in values]
        in_range = (
            (lambda prop: 0.0 <= prop <= 1.0) if include_axis else (lambda prop: 0.0 <= prop < 1.0)
        )
        return [
            self.position.y + prop * self.length if in_range(prop) else None
            for prop in proportions_of_range
        ]



from abc import abstractmethod

from .helpers import collapse_element_list
from .shapes import Shape, Line, Text, Rect, Circle
from .shared import number


class Legend(Shape):
    """
    base class for legends
    """

    @abstractmethod
    def get_element_list(self) -> list: ...


class LineLegend(Legend):
    """
    legend for line charts
    """

    default_line_legend_text_styles = {"alignment-baseline": "middle"}

    def __init__(
            self,
            x: number,
            y: number,
            series,
            element_x: number,
            element_y: number,
            line_length: number,
            line_text_gap: number,
    ):
        super().__init__(x, y)
        self.series = series
        self.lines, self.texts = [], []
        x_pos, y_pos = self.position.x, self.position.y
        for index, series in enumerate(self.series):
            self.lines.append(
                Line(x_pos, y_pos, width=line_length, height=0, styles=self.series[series].styles)
            )
            self.texts.append(
                Text(
                    x_pos + line_length + line_text_gap,
                    y_pos,
                    content=series,
                    styles=self.default_line_legend_text_styles,
                )
            )
            x_pos += element_x
            y_pos += element_y

    def get_element_list(self) -> list[str]:
        return collapse_element_list(self.lines, self.texts)


class BarLegend(Legend):
    """
    legend for bar charts
    """

    default_line_legend_text_styles = {"alignment-baseline": "middle"}

    def __init__(
            self,
            x: number,
            y: number,
            series,
            element_x: number,
            element_y: number,
            bar_width: number,
            bar_height: number,
            bar_text_gap: number,
    ):
        super().__init__(x, y)
        self.series = series
        self.lines, self.texts = [], []
        x_pos, y_pos = self.position.x, self.position.y
        for index, series in enumerate(self.series):
            self.lines.append(
                Rect(
                    x_pos,
                    y_pos - bar_height / 2,
                    width=bar_width,
                    height=bar_height,
                    styles=self.series[series].styles,
                ),
            )
            self.texts.append(
                Text(
                    x_pos + bar_width + bar_text_gap,
                    y_pos,
                    content=series,
                    styles=self.default_line_legend_text_styles,
                ),
            )
            x_pos += element_x
            y_pos += element_y

    def get_element_list(self) -> list[str]:
        return collapse_element_list(self.lines, self.texts)


class ScatterLegend(Legend):
    """
    legend for a scatter chart
    """

    default_scatter_legend_text_styles = {"alignment-baseline": "middle"}

    def __init__(
            self,
            x: number,
            y: number,
            series,
            element_x: number,
            element_y: number,
            shape_text_gap: number,
    ):
        super().__init__(x, y)
        self.series = series
        self.legends, self.texts = [], []
        x_pos, y_pos = self.position.x, self.position.y
        for index, series in enumerate(self.series):
            self.legends.append(
                self.series[series].shape_template(x_pos, y_pos, styles=self.series[series].styles)
            )
            self.texts.append(
                Text(
                    x_pos + shape_text_gap,
                    y_pos,
                    content=series,
                    styles=self.default_scatter_legend_text_styles,
                )
            )
            x_pos += element_x
            y_pos += element_y

    def get_element_list(self) -> list[str]:
        return collapse_element_list(self.legends, self.texts)


class DonutLegend(Legend):
    """
    Legend for donut (or pie) charts using small colored circles.
    """

    default_circle_legend_text_styles = {"alignment-baseline": "middle"}

    def __init__(
            self,
            x: number,
            y: number,
            series,
            element_x: number,
            element_y: number,
            circle_radius: number,
            circle_text_gap: number,
    ):
        super().__init__(x, y)
        self.series = series
        self.circles, self.texts = [], []
        x_pos, y_pos = self.position.x, self.position.y
        for index, series_name in enumerate(self.series):
            self.circles.append(
                Circle(
                    x=x_pos + circle_radius,
                    y=y_pos,
                    radius=circle_radius,
                    styles=self.series[series_name].styles,
                )
            )
            self.texts.append(
                Text(
                    x_pos + 2 * circle_radius + circle_text_gap,
                    y_pos,
                    content=series_name,
                    styles=self.default_circle_legend_text_styles,
                )
            )
            x_pos += element_x
            y_pos += element_y

    def get_element_list(self) -> list[str]:
        return collapse_element_list(self.circles, self.texts)



from abc import ABC, abstractmethod
from collections.abc import Callable
from itertools import zip_longest, cycle
from typing import Any

from .axes import Axis, XAxis, YAxis, CategoryYAxis
from .helpers import collapse_element_list, default_format
from .legends import BarLegend, Legend, LineLegend, ScatterLegend, DonutLegend
from .scales import make_categories_scale, make_logarithmic_scale, make_linear_scale
from .series import BarSeries, DonutSegment, LineSeries, ScatterSeries, Series
from .shapes import Circle, Group, Line, Point
from .shared import named_styles, number, numbers_sequence, style_def
from .styles import render_all_styles


def no_series_constructor(
        x_values: list | tuple,
        y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
        x_axis: Axis,
        y_axis: Axis,
        series_names: list[str],
        bar_width: number,
        bar_gap: number,
) -> dict[str, Series]:
    _ignore = x_axis, y_axis, bar_width, bar_gap
    if len(y_values) != len(series_names):
        raise ValueError("y_values and series_names must have the same length")
    if not all(len(y_value) == len(x_values) for y_value in y_values):
        raise ValueError("y_values must all have the same length as x_values")
    return {name: Series(x_values[0], y_value[0]) for name, y_value in zip(series_names, y_values)}


def line_series_constructor(
        x_values: list | tuple,
        y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
        x_axis: Axis,
        y_axis: Axis,
        series_names: list[str],
        bar_width: number,
        bar_gap: number,
) -> dict[str, Series]:
    _ignore = bar_width, bar_gap
    if len(y_values) != len(series_names):
        raise ValueError("y_values and series_names must have the same length")
    if not all(len(y_value) == len(x_values) for y_value in y_values):
        raise ValueError("y_values must all have the same length as x_values")
    return {
        name: LineSeries(
            points=[
                Point(x=x, y=y)  # type: ignore[arg-type]
                for x, y in zip(x_axis.get_positions(x_values), y_axis.get_positions(y_value))
            ],
            x_values=x_values,
            y_values=y_value,  # type: ignore[arg-type]
        )
        for name, y_value in zip(series_names, y_values)
    }


def bar_series_constructor(
        x_values: list | tuple,
        y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
        x_axis: Axis,
        y_axis: Axis,
        series_names: list[str],
        bar_width: number,
        bar_gap: number,
) -> dict[str, Series]:
    if len(y_values) != len(series_names):
        raise ValueError("y_values and series_names must have the same length")
    if not all(len(y_value) == len(x_values) for y_value in y_values):
        raise ValueError("y_values must all have the same length as x_values")
    no_series = len(series_names)
    bar_span = bar_width + bar_gap
    bar_shift = bar_span * (no_series - 1) / 2
    return {
        name: BarSeries(
            points=[
                Point(x=x + bar_nr * bar_span - bar_shift, y=y)  # type: ignore[arg-type, operator]
                for x, y in zip(x_axis.get_positions(x_values), y_axis.get_positions(y_value))
            ],
            x_values=x_values,
            y_values=y_value,  # type: ignore[arg-type]
            bar_heights=[
                y_axis.position.y + y_axis.length - y if y is not None else 0
                for y in y_axis.get_positions(y_value)
            ],
            bar_width=bar_width,
        )
        for bar_nr, name, y_value in zip(range(no_series), series_names, y_values)
    }


def normalised_bar_series_constructor(
        x_values: list | tuple,
        y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
        x_axis: Axis,
        y_axis: Axis,
        series_names: list[str],
        bar_width: number,
        bar_gap: number,
) -> dict[str, Series]:
    _ignore = bar_gap
    if len(y_values) < 1:
        raise ValueError("y_values should not be empty")
    if len(y_values) != len(series_names):
        raise ValueError("y_values and series_names must have the same length")
    if not all(len(y_value) == len(x_values) for y_value in y_values):
        raise ValueError("y_values must all have the same length as x_values")
    rtn: dict[str, Series] = {}
    prev_cumulative_scaled_y_values = [0] * len(y_values[0])
    total_values = [sum(y) for y in zip(*y_values)]
    x_positions = x_axis.get_positions(x_values)
    for name, y_value in zip(series_names, y_values):
        cumulative_scaled_y_values = [
            min(a + b / t, 1.0) if t != 0 else a
            for a, b, t in zip(prev_cumulative_scaled_y_values, y_value, total_values)
        ]
        prev_scaled_positions = y_axis.get_positions(prev_cumulative_scaled_y_values)
        scaled_positions = y_axis.get_positions(cumulative_scaled_y_values)
        rtn[name] = BarSeries(
            points=[Point(x=x, y=y) for x, y in zip(x_positions, scaled_positions)],  # type: ignore[arg-type]
            x_values=x_values,
            y_values=y_value,  # type: ignore[arg-type]
            bar_heights=[b - a for a, b in zip(scaled_positions, prev_scaled_positions)],  # type: ignore[operator]
            bar_width=bar_width,
        )
        prev_cumulative_scaled_y_values = cumulative_scaled_y_values
    return rtn


def scatter_series_constructor(
        x_values: list | tuple,
        y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
        x_axis: Axis,
        y_axis: Axis,
        series_names: list[str],
        bar_width: number,
        bar_gap: number,
) -> dict[str, Series]:
    _ignore = bar_width, bar_gap
    if len(y_values) != len(series_names):
        raise ValueError("y_values and series_names must have the same length")
    if not all(len(y_value) == len(x_values) for y_value in y_values):
        raise ValueError("y_values must all have the same length as x_values")
    return {
        name: ScatterSeries(
            points=[
                Point(x=x, y=y)  # type: ignore[arg-type]
                for x, y in zip(x_axis.get_positions(x_values), y_axis.get_positions(y_value))
            ],
            x_values=x_values,
            y_values=y_value,  # type: ignore[arg-type]
        )
        for name, y_value in zip(series_names, y_values)
    }


def horizontal_bar_series_constructor(
        y_values: list | tuple,
        x_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
        y_axis: Axis,
        x_axis: Axis,
        series_names: list[str],
        bar_width: number,
        bar_gap: number,
) -> dict[str, Series]:
    """
    Constructor for horizontal bar series.
    In horizontal bars:
      - y_values are categories (shown on y-axis/vertical)
      - x_values are the numerical values (shown on x-axis/horizontal)
      - bars grow horizontally (left to right)
    """
    if len(x_values) != len(series_names):
        raise ValueError("x_values and series_names must have the same length")
    if not all(len(x_value) == len(y_values) for x_value in x_values):
        raise ValueError("x_values must all have the same length as y_values")
    no_series = len(series_names)
    bar_span = bar_width + bar_gap
    bar_shift = bar_span * (no_series - 1) / 2
    from .series import HorizontalBarSeries
    return {
        name: HorizontalBarSeries(
            points=[
                Point(x=x_axis.position.x, y=y + bar_nr * bar_span - bar_shift)  # type: ignore[arg-type, operator]
                for y in y_axis.get_positions(y_values)
            ],
            x_values=y_values,
            y_values=x_value,  # type: ignore[arg-type]
            bar_heights=[
                x - x_axis.position.x if x is not None else 0
                for x in x_axis.get_positions(x_value)
            ],
            bar_width=bar_width,
        )
        for bar_nr, name, x_value in zip(range(no_series), series_names, x_values)
    }


def default_x_range_constructor(x_values: list | tuple) -> list:
    return [v for v in x_values]


def default_y_range_constructor(
        y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
) -> list:
    return [v for series in y_values for v in series]


class Chart(ABC):
    """
    overall svg template for chart
    """

    svg_begin_template = '<svg viewBox="0 0 {width} {height}" xmlns="http://www.w3.org/2000/svg">'

    height: number
    width: number
    custom_elements: list[str]
    series: dict[str, Any]

    def __init__(self, height: number, width: number) -> None:
        self.height = height
        self.width = width
        self.custom_elements = []
        self.series = {}

    @abstractmethod
    def get_element_list(self) -> list[str]: ...

    def add_custom_element(self, custom_element) -> None:
        self.custom_elements.append(custom_element)

    def modify_series(self, modifier: Callable) -> None:
        self.series = {name: modifier(series) for name, series in self.series.items()}

    def render(self) -> str:
        return "\n".join(
            [
                self.svg_begin_template.format(height=self.height, width=self.width),
                *self.get_element_list(),
                "</svg>",
            ]
        )

    def render_with_all_styles(
            self,
            styles: named_styles | None = None,
            include_default: bool = True,
    ) -> str:
        """
        :param styles: styles to use
        :param include_default: also use the default styles (to enable things like hover text)
        :return:
        """
        return "\n".join(
            [
                self.svg_begin_template.format(height=self.height, width=self.width),
                "<style>",
                render_all_styles(styles, include_default),
                "</style>",
                *self.get_element_list(),
                "</svg>",
            ]
        )

    def save(
            self,
            file_path: str,
            styles: named_styles | None = None,
            include_default: bool = True
    ) -> None:
        """
        Save the chart to a file.
        :param file_path: file path to write to
        :param styles: styles to use (if provided, will render with all styles)
        :param include_default: also use the default styles (only used when styles is provided)
        :return:
        """
        with open(file_path, "w+") as file:
            if styles is not None or not include_default:
                file.write(self.render_with_all_styles(styles, include_default))
            else:
                file.write(self.render())

    @staticmethod
    def generate_series_names(
            prefix: str,
            n: int,
            names: list[str] | tuple[str, ...] | None,
    ) -> list[str]:
        """
        generate missing names for series
        """
        return [
                   real if real is not None else generated
                   for real, generated in zip_longest(
                names if names is not None else [],
                [f"{prefix} {k}" for k in range(1, n + 1)],
            )
               ][:n]


class CartesianChart(Chart):
    """
    Base class for charts with two perpendicular axes (X and Y).
    This class provides the foundation for both vertical and horizontal oriented charts.
    Subclasses should define orientation-specific behavior.
    """

    __colour_defaults__ = ["green", "red", "blue", "orange", "yellow", "black"]

    default_major_grid_styles = {"stroke": "#6e6e6e", "stroke-width": "0.6"}
    default_minor_grid_styles = {"stroke": "#6e6e6e", "stroke-width": "0.2"}
    colour_property = "stroke"

    # The defaults are for axis classes - subclasses should override
    x_axis_type = Axis
    y_axis_type = Axis
    x_axis_scale_maker = staticmethod(make_linear_scale)
    y_axis_scale_maker = staticmethod(make_linear_scale)

    # The defaults are for series - subclasses should override
    x_range_constructor = staticmethod(default_x_range_constructor)
    y_range_constructor = staticmethod(default_y_range_constructor)
    series_constructor = staticmethod(no_series_constructor)

    def set_palette(self, colours: list[str] | tuple[str, ...]) -> None:
        for series, colour in zip(self.series, cycle(colours)):
            self.series[series].styles[self.colour_property] = colour


class VerticalChart(CartesianChart):
    """
    Any chart with a vertical y-axis and a horizontal x-axis
    - all lines share the same x values
    - y values differ
    """

    def __init__(
            self,
            # chart data
            x_values: list | tuple,
            y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
            sec_y_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...] | None = None,
            y_names: list[str] | None = None,
            sec_y_names: list[str] | None = None,
            # x-axis
            x_min: Any = None,
            x_max: Any = None,
            x_zero: bool = False,
            x_max_ticks: int = 12,
            x_shift: bool = False,
            x_label_format: Callable = default_format,
            x_axis_title: str | None = None,
            x_axis_title_styles: dict | None = None,
            x_axis_title_offset: int = 40,
            # primary y-axis
            y_min: Any = None,
            y_max: Any = None,
            y_zero: bool = False,
            y_max_ticks: int = 12,
            y_shift: bool = False,
            y_label_format: Callable = default_format,
            y_axis_title: str | None = None,
            y_axis_title_styles: dict | None = None,
            y_axis_title_offset: int = 40,
            # secondary y-axis
            sec_y_min: Any = None,
            sec_y_max: Any = None,
            sec_y_zero: bool = False,
            sec_y_max_ticks: int = 12,
            sec_y_shift: bool = False,
            sec_y_label_format: Callable = default_format,
            sec_y_axis_title: str | None = None,
            sec_y_axis_title_styles: dict | None = None,
            sec_y_axis_title_offset: int = 40,
            # canvas
            left_margin: number = 100,
            right_margin: number = 100,
            y_margin: number = 100,
            height: number = 600,
            width: number = 800,
            bar_width: number = 40,
            bar_gap: number = 2,
            colours: list[str] | tuple[str, ...] | None = None,
    ):
        """
        create a simple line chart
        :param x_values: the list of x values shared by all lines
        :param y_values: a list line values for the primary y-axis, each a list itself
        :param sec_y_values:  a list line values for the secondary y-axis, each a list itself
        :param y_names: optional list of names of the lines of the primary y-axis
        :param sec_y_names: optional list of names of the lines of the secondary y-axis
        :param x_min: optional minimum x value, only used in numeric axis
        :param x_max: optional maximum x value, only used in numeric axis
        :param x_zero: optionally force 0 to be included on the x-axis
        :param x_max_ticks: optional maximum number of ticks on the x-axis
        :param x_shift: optionally shift the x-axis - True: left side of graph touches the y-axis, value: shift graph left by that amount
        :param x_label_format: optional format of labels on the x-axis
        :param y_min: optional minimum value on the primary y-axis if it is numeric
        :param y_max: optional maximum value on the primary y-axis if is is numeric
        :param y_zero: optionally force 0 to be included on the primary y-axis
        :param y_max_ticks: optional maximum number of ticks on the primary y-axis
        :param y_shift: optionally shift the y-axis - True: bottom side of graph touches the x-axis, value: shift graph down by that amount
        :param y_label_format: optional format of labels on the primary y-axis
        :param sec_y_min: optional minimum value on the secondary y-axis
        :param sec_y_max: optional maximum value on the secondary y-axis
        :param sec_y_zero: optionally force 0 to be included on the secondary y-axis
        :param sec_y_max_ticks: optional maximum number of ticks on the secondary y-axis
        :param sec_y_shift: optionally shift the secondary y-axis - True: bottom side of graph touches the x-axis, value: shift graph down by that amount
        :param sec_y_label_format: optional format of labels on the secondary y-axis
        :param left_margin: optional left margin for the x-axis
        :param right_margin: optional right margin for the x-axis
        :param y_margin: optional margin for the y-axis
        :param height: optional height of the graph
        :param width: optional width of the graph
        :param colours: optional list of colours for the series
        """
        super().__init__(height, width)
        self.x_axis = self.x_axis_type(  # type: ignore[abstract]
            x_position=left_margin,
            y_position=height - y_margin,
            data_points=self.x_range_constructor(x_values),
            axis_length=width - left_margin - right_margin,
            label_format=x_label_format,
            max_ticks=x_max_ticks,
            min_value=x_min,
            max_value=x_max,
            include_zero=x_zero,
            shift=x_shift,
            scale_maker=self.x_axis_scale_maker,
            title=x_axis_title,
            title_styles=x_axis_title_styles,
            title_offset=x_axis_title_offset
        )
        self.y_axis = self.y_axis_type(  # type: ignore[abstract]
            x_position=left_margin,
            y_position=y_margin,
            data_points=self.y_range_constructor(y_values),
            axis_length=height - 2 * y_margin,
            label_format=y_label_format,
            max_ticks=y_max_ticks,
            min_value=y_min,
            max_value=y_max,
            include_zero=y_zero,
            shift=y_shift,
            scale_maker=self.y_axis_scale_maker,
            secondary=False,
            title=y_axis_title,
            title_offset=y_axis_title_offset,
            title_styles=y_axis_title_styles
        )
        series_names = self.generate_series_names("Series", len(y_values), y_names)
        self.series = self.series_constructor(
            x_values,
            y_values,
            self.x_axis,
            self.y_axis,
            series_names,
            bar_width,
            bar_gap,
        )

        if sec_y_values is None:
            self.sec_y_axis = None
        else:
            sec_series_names = self.generate_series_names(
                "Secondary series",
                len(sec_y_values),
                sec_y_names,
            )
            self.sec_y_axis = self.y_axis_type(  # type: ignore[abstract]
                x_position=width - right_margin,
                y_position=y_margin,
                data_points=default_y_range_constructor(sec_y_values),
                axis_length=height - 2 * y_margin,
                label_format=sec_y_label_format,
                max_ticks=sec_y_max_ticks,
                min_value=sec_y_min,
                max_value=sec_y_max,
                include_zero=sec_y_zero,
                shift=sec_y_shift,
                scale_maker=self.y_axis_scale_maker,
                secondary=True,
                title=sec_y_axis_title,
                title_styles=sec_y_axis_title_styles,
                title_offset=sec_y_axis_title_offset
            )
            self.series.update(
                self.series_constructor(
                    x_values,
                    sec_y_values,
                    self.x_axis,
                    self.sec_y_axis,
                    sec_series_names,
                    bar_width,
                    bar_gap,
                )
            )
        self.legend: Legend | None = None
        self.set_palette(colours if colours else self.__colour_defaults__)

    def add_legend(
            self,
            x_position: number = 730,
            y_position: number = 200,
            element_x: number = 0,
            element_y: number = 20,
            line_length: number = 20,
            line_text_gap: number = 5,
            **kwargs,
    ):
        self.legend = LineLegend(
            x_position,
            y_position,
            self.series,
            element_x,
            element_y,
            line_length,
            line_text_gap,
        )

    def add_grids(
            self,
            minor_x_ticks: int = 0,
            minor_y_ticks: int = 0,
            major_grid_style: style_def | None = None,
            minor_grid_style: style_def | None = None,
    ):
        self.add_y_grid(minor_y_ticks, major_grid_style, minor_grid_style)
        self.add_x_grid(minor_x_ticks, major_grid_style, minor_grid_style)

    def add_y_grid(
            self,
            minor_ticks: int = 0,
            major_grid_style: style_def | None = None,
            minor_grid_style: style_def | None = None,
    ):
        major_style = (
            major_grid_style.copy()
            if major_grid_style is not None
            else self.default_major_grid_styles.copy()
        )
        minor_style = (
            minor_grid_style.copy()
            if minor_grid_style is not None
            else self.default_minor_grid_styles.copy()
        )
        positions = self.x_axis.get_positions(self.x_axis.scale.ticks, include_axis=False)
        for pos in positions:
            if pos is None:  # shifted out of the visible range
                continue
            minor_unit = self.x_axis.length / (len(self.x_axis.scale.ticks) - 1) / (minor_ticks + 1)
            for grid_line_nr in range(minor_ticks + 1):  # 0: major, others: minor
                self.y_axis.grid_lines.append(
                    Line(
                        x=pos - grid_line_nr * minor_unit,
                        y=self.x_axis.position.y - self.y_axis.length,
                        width=0,
                        height=self.y_axis.length,
                        styles=major_style if grid_line_nr == 0 else minor_style,
                    )
                )

    def add_x_grid(
            self,
            minor_ticks: int = 0,
            major_grid_style: style_def | None = None,
            minor_grid_style: style_def | None = None,
    ):
        major_style = (
            major_grid_style.copy()
            if major_grid_style is not None
            else self.default_major_grid_styles.copy()
        )
        minor_style = (
            minor_grid_style.copy()
            if minor_grid_style is not None
            else self.default_minor_grid_styles.copy()
        )
        positions = self.y_axis.get_positions(self.y_axis.scale.ticks, include_axis=False)
        for pos in positions:
            if pos is None:  # shifted out of the visible range
                continue
            minor_unit = self.y_axis.length / (len(self.y_axis.scale.ticks) - 1) / (minor_ticks + 1)
            for grid_line_nr in range(minor_ticks + 1):  # 0: major, others: minor
                self.y_axis.grid_lines.append(
                    Line(
                        x=self.y_axis.position.x,
                        y=pos + grid_line_nr * minor_unit,
                        width=self.x_axis.length,
                        height=0,
                        styles=major_style if grid_line_nr == 0 else minor_style,
                    )
                )

    def add_hover_modifier(
            self,
            modifier: Callable,
            radius: number,
            series_list: list[str] | tuple[str, ...] | None = None,
    ):
        def build_hover_marker(point, x_value, y_value, series_name):
            series_styles = self.series[series_name].styles
            circle = Circle(point.x, y=point.y, radius=radius, styles={"style": "opacity:0;"})
            mod = modifier(
                point,
                x_value=x_value,
                y_value=y_value,
                series_name=series_name,
                styles=series_styles,
            )
            return Group(children=[circle] + mod, classes=["psc-hover-group"])

        series_list = [s for s in self.series] if series_list is None else series_list
        for s in self.series:
            if s in series_list:
                hover_markers = [
                    build_hover_marker(p, x, y, s)
                    for p, x, y in self.series[s].pv_generator
                    if x is not None and y is not None and p.x is not None and p.y is not None
                ]
                self.series[s].add_custom_elements(hover_markers)

    def get_element_list(self) -> list[str]:
        return collapse_element_list(
            [self.x_axis],
            [self.y_axis],
            [self.legend],
            [self.sec_y_axis],
            [self.series[s] for s in self.series],
            self.custom_elements,
        )


class HorizontalChart(CartesianChart):
    """
    Chart with a horizontal value axis and vertical category/data axis.
    Useful for horizontal bar charts where categories are on Y-axis and values on X-axis.
    """

    def __init__(
            self,
            # chart data
            x_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...],
            y_values: list | tuple,
            sec_x_values: list[list] | list[tuple] | tuple[list, ...] | tuple[tuple, ...] | None = None,
            x_names: list[str] | None = None,
            sec_x_names: list[str] | None = None,
            # y-axis (categories/vertical)
            y_min: Any = None,
            y_max: Any = None,
            y_zero: bool = False,
            y_max_ticks: int = 12,
            y_shift: bool = False,
            y_label_format: Callable = default_format,
            y_axis_title: str | None = None,
            y_axis_title_styles: dict | None = None,
            y_axis_title_offset: int = 40,
            # primary x-axis (values/horizontal)
            x_min: Any = None,
            x_max: Any = None,
            x_zero: bool = False,
            x_max_ticks: int = 12,
            x_shift: bool = False,
            x_label_format: Callable = default_format,
            x_axis_title: str | None = None,
            x_axis_title_styles: dict | None = None,
            x_axis_title_offset: int = 40,
            # secondary x-axis
            sec_x_min: Any = None,
            sec_x_max: Any = None,
            sec_x_zero: bool = False,
            sec_x_max_ticks: int = 12,
            sec_x_shift: bool = False,
            sec_x_label_format: Callable = default_format,
            sec_x_axis_title: str | None = None,
            sec_x_axis_title_styles: dict | None = None,
            sec_x_axis_title_offset: int = 40,
            # canvas
            left_margin: number = 100,
            right_margin: number = 100,
            x_margin: number = 100,
            height: number = 600,
            width: number = 800,
            bar_width: number = 40,
            bar_gap: number = 2,
            colours: list[str] | tuple[str, ...] | None = None,
    ):
        """
        Create a horizontal chart where categories are on Y-axis (vertical) and values on X-axis (horizontal).
        Parameters match the physical axes: x_values are horizontal (values), y_values are vertical (categories).
        """
        super().__init__(height, width)

        # In horizontal charts:
        # - Y-axis is vertical and shows categories (y_values)
        # - X-axis is horizontal and shows values (x_values)
        self.y_axis = self.y_axis_type(  # type: ignore[abstract]
            x_position=left_margin,
            y_position=x_margin,
            data_points=self.x_range_constructor(y_values),
            axis_length=height - 2 * x_margin,
            label_format=y_label_format,
            max_ticks=y_max_ticks,
            min_value=y_min,
            max_value=y_max,
            include_zero=y_zero,
            shift=y_shift,
            scale_maker=self.y_axis_scale_maker,
            secondary=False,
            title=y_axis_title,
            title_styles=y_axis_title_styles,
            title_offset=y_axis_title_offset
        )

        self.x_axis = self.x_axis_type(  # type: ignore[abstract]
            x_position=left_margin,
            y_position=height - x_margin,
            data_points=self.y_range_constructor(x_values),
            axis_length=width - left_margin - right_margin,
            label_format=x_label_format,
            max_ticks=x_max_ticks,
            min_value=x_min,
            max_value=x_max,
            include_zero=x_zero,
            shift=x_shift,
            scale_maker=self.x_axis_scale_maker,
            title=x_axis_title,
            title_styles=x_axis_title_styles,
            title_offset=x_axis_title_offset
        )

        series_names = self.generate_series_names("Series", len(x_values), x_names)
        self.series = self.series_constructor(
            y_values,
            x_values,
            self.y_axis,
            self.x_axis,
            series_names,
            bar_width,
            bar_gap,
        )

        self.sec_x_axis = None
        if sec_x_values is not None:
            sec_series_names = self.generate_series_names(
                "Secondary series",
                len(sec_x_values),
                sec_x_names,
            )
            self.sec_x_axis = self.x_axis_type(  # type: ignore[abstract]
                x_position=left_margin,
                y_position=x_margin,
                data_points=default_x_range_constructor(sec_x_values),
                axis_length=width - left_margin - right_margin,
                label_format=sec_x_label_format,
                max_ticks=sec_x_max_ticks,
                min_value=sec_x_min,
                max_value=sec_x_max,
                include_zero=sec_x_zero,
                shift=sec_x_shift,
                scale_maker=self.x_axis_scale_maker,
                secondary=True,
                title=sec_x_axis_title,
                title_styles=sec_x_axis_title_styles,
                title_offset=sec_x_axis_title_offset
            )
            self.series.update(
                self.series_constructor(
                    x_values,
                    sec_x_values,
                    self.y_axis,
                    self.sec_x_axis,
                    sec_series_names,
                    bar_width,
                    bar_gap,
                )
            )

        self.legend: Legend | None = None
        self.set_palette(colours if colours else self.__colour_defaults__)

    def get_element_list(self) -> list[str]:
        return collapse_element_list(
            [self.x_axis],
            [self.y_axis],
            [self.legend],
            [self.sec_x_axis],
            [self.series[s] for s in self.series],
            self.custom_elements,
        )


class LineChart(VerticalChart):
    x_axis_type = XAxis  # type: ignore[assignment]
    y_axis_type = YAxis  # type: ignore[assignment]
    series_constructor = staticmethod(line_series_constructor)

    def __init__(self, *args, **kwargs):
        """
        intercept init to handle optional logarithmic scale
        :param x_log: optionally enable logarithmic scale
        :param y_log: optionally enable logarithmic scale
        """
        x_log = kwargs.pop("x_log", False)
        y_log = kwargs.pop("y_log", False)
        if x_log:
            self.x_axis_scale_maker = staticmethod(make_logarithmic_scale)
        if y_log:
            self.y_axis_scale_maker = staticmethod(make_logarithmic_scale)
        super().__init__(*args, **kwargs)


class SimpleLineChart(LineChart):
    x_axis_type = XAxis
    y_axis_type = YAxis
    series_constructor = staticmethod(line_series_constructor)


class BarChart(LineChart):
    x_axis_type = XAxis
    y_axis_type = YAxis
    x_axis_scale_maker = staticmethod(make_categories_scale)
    y_axis_scale_maker = staticmethod(make_linear_scale)
    series_constructor = staticmethod(bar_series_constructor)
    colour_property = "fill"

    def add_legend(  # type: ignore[override]
            self,
            x_position: number = 730,
            y_position: number = 200,
            element_x: number = 0,
            element_y: number = 20,
            bar_width: number = 30,
            bar_height: number = 5,
            bar_text_gap: number = 5,
            **kwargs,
    ):
        self.legend = BarLegend(
            x_position,
            y_position,
            self.series,
            element_x,
            element_y,
            bar_width,
            bar_height,
            bar_text_gap,
        )


class HorizontalBarChart(HorizontalChart):
    """
    Horizontal bar chart where categories are on Y-axis (vertical) and values on X-axis (horizontal).
    Bars grow from left to right.
    """
    x_axis_type = XAxis
    y_axis_type = CategoryYAxis
    x_axis_scale_maker = staticmethod(make_linear_scale)
    y_axis_scale_maker = staticmethod(make_categories_scale)
    series_constructor = staticmethod(horizontal_bar_series_constructor)
    colour_property = "fill"

    def add_legend(  # type: ignore[override]
            self,
            x_position: number = 730,
            y_position: number = 200,
            element_x: number = 0,
            element_y: number = 20,
            bar_width: number = 30,
            bar_height: number = 5,
            bar_text_gap: number = 5,
            **kwargs,
    ):
        self.legend = BarLegend(
            x_position,
            y_position,
            self.series,
            element_x,
            element_y,
            bar_width,
            bar_height,
            bar_text_gap,
        )


class NormalisedBarChart(LineChart):
    x_axis_type = XAxis
    y_axis_type = YAxis
    x_axis_scale_maker = staticmethod(make_categories_scale)
    y_axis_scale_maker = staticmethod(make_linear_scale)
    series_constructor = staticmethod(normalised_bar_series_constructor)
    y_range_constructor = staticmethod(lambda y_values: [0, 1])
    colour_property = "fill"

    def add_legend(  # type: ignore[override]
            self,
            x_position: number = 730,
            y_position: number = 200,
            element_x: number = 0,
            element_y: number = 20,
            bar_width: number = 30,
            bar_height: number = 5,
            bar_text_gap: number = 5,
            **kwargs,
    ):
        self.legend = BarLegend(
            x_position,
            y_position,
            self.series,
            element_x,
            element_y,
            bar_width,
            bar_height,
            bar_text_gap,
        )


class ScatterChart(LineChart):
    x_axis_type = XAxis
    y_axis_type = YAxis
    series_constructor = staticmethod(scatter_series_constructor)
    colour_property = "fill"

    def add_legend(  # type: ignore[override]
            self,
            x_position: number = 730,
            y_position: number = 200,
            element_x: number = 0,
            element_y: number = 20,
            shape_text_gap: number = 5,
            **kwargs,
    ):
        self.legend = ScatterLegend(
            x_position,
            y_position,
            self.series,
            element_x,
            element_y,
            shape_text_gap,
        )


class DonutChart(Chart):
    """
    A donut style chart which is similar to a pie chart but has a blank interior
    """

    __segment_colour_defaults__ = ["green", "red", "blue", "orange", "yellow", "black"]

    def __init__(
            self,
            values: numbers_sequence,
            labels: list[str] | tuple[str, ...] | None = None,
            colours: list[str] | tuple[str, ...] | None = None,
            height: number = 400,
            width: number = 400,
            centre_x: number = 150,
            centre_y: number = 200,
            radius_inner: number = 55,
            radius_outer: number = 150,
            rotation: number = 70,
    ):
        """
        create a donut chart
        :param values: values to chart
        :param labels: labels to each segment
        :param height: canvas height
        :param width: canvas width
        :param centre_x: horizontal centre of donut
        :param centre_y: vertical centre of donut
        :param radius_inner: inner radius of donut (blank area)
        :param radius_outer: outer radius of donut (other area)
        :param rotation: rotation offset
        """
        super().__init__(height, width)
        self.values = values
        series_names = self.generate_series_names("Series", len(values), labels)
        # compute start and end angles for the value segments
        accumulated_values: list[number] = [0]
        for value in values:
            accumulated_values.append(value + accumulated_values[-1])
        total_value = accumulated_values[-1]
        rotated_angles = [rotation + (360 * value) / total_value for value in accumulated_values]
        start_end_angles = [
            rotated_angles[index: index + 2] for index in range(len(rotated_angles) - 1)
        ]
        self.colours = colours if colours is not None else self.__segment_colour_defaults__
        self.labels = None
        # create value segments
        self.legend: Legend | None = None
        for index, (start_theta, end_theta), name in zip(
                range(len(values)),
                start_end_angles,
                series_names,
        ):
            colour = self.colours[index % len(self.colours)]
            self.series[name] = DonutSegment(
                colour,
                start_theta,
                end_theta,
                radius_inner,
                radius_outer,
                centre_x,
                centre_y,
            )

    def add_hover_modifier(self, modifier: Callable):
        names = list(self.series)
        segments = [self.series[name] for name in names]
        chart_total = sum(self.values)
        self.series = {
            n: Group(
                children=(
                        [s] + modifier(position=s.position, name=n, value=v, chart_total=chart_total)
                )
            )
            for n, v, s in zip(names, self.values, segments)
        }
        for s in self.series:
            self.series[s].add_classes(["psc-hover-group"])

    def add_legend(  # type: ignore[override]
            self,
            x_position: number = 320,
            y_position: number = 175,
            element_x: number = 0,
            element_y: number = 20,
            circle_radius: number = 5,
            circle_text_gap: number = 5,
            **kwargs,
    ):
        self.legend = DonutLegend(
            x_position,
            y_position,
            self.series,
            element_x,
            element_y,
            circle_radius,
            circle_text_gap,
        )

    def get_element_list(self) -> list[str]:
        return collapse_element_list([self.series[s] for s in self.series], self.custom_elements, [self.legend])



"""
Simple test script to demonstrate the new HorizontalBarChart
"""
import pysvgchart as psc

# Create a simple horizontal bar chart
categories = ['Apples', 'Bananas', 'Cherries', 'Dates', 'Elderberries']
values = [[25, 40, 15, 30, 35]]

horizontal_bar_chart = psc.HorizontalBarChart(
    x_values=values,
    y_values=categories,
    x_names=['Sales'],
    x_zero=True,
    y_axis_title='Products',
    x_axis_title='Units Sold',
    width=800,
    height=600,
    left_margin=200,
    y_axis_title_offset=90,
)

# Save the chart
horizontal_bar_chart.save('showcase/horizontal_bar_test.svg')


# Also test that existing VerticalChart still works
vertical_bar_chart = psc.BarChart(
    x_values=categories,
    y_values=values,
    y_names=['Sales'],
    y_zero=True,
    x_axis_title='Products',
    y_axis_title='Units Sold',
    width=800,
    height=400,
)

vertical_bar_chart.save('showcase/vertical_bar_test.svg')

